/*
 * USB.c by Aidan Mocke (29/12/2020)
 * ---------------------------------
 * Contains basic USB Endpoint 0 setup packet handling
 * Calls USB_EP1_tasks() and USB_EP2_tasks() on receipt of data on Endpoint 1 and Endpoint 2 respectively (these functions must be defined elsewhere)
 */
#include "USB.h"
#include <xc.h>
#include <string.h>

#define	STR_DESC(l)	(((((l))<<1) & 0xFF) | 0x0300)
// String descriptors. All have a common format: [byte] string_length, [byte] data_type (3 = string), UTF-16 encoded string, hence all the 0's between each 8-byte character

// Language description
unsigned char string0[] =  { 4, 0x03, 0x09, 0x04}; // 0x0409 = English

// Vendor description
unsigned short string1[] = { STR_DESC(11),
                            'A','i','d','a','n',' ','M','o','c','k','e'};   

// Product description
unsigned short string2[] = { STR_DESC(12),
                            'U','S','B',' ','M','S','D',' ','T','e','s','t'};
 
// Serial number description
unsigned short string3[] = { STR_DESC(7),
                            '1','2','3','-','3','2','1'};

// Configuration description
unsigned short string4[] = { STR_DESC(8),
                            'U','L','T','R','A',' ','S','E'};
// Interface description 
unsigned short string5[] = { STR_DESC(7),
                            'S','t','o','r','a','g','e'};

volatile char USB_set_address = 0;
volatile char USB_line_control_state = 0;
unsigned char USB_address = 0;
USB_TRANSACTION USB_transaction;
USB_LINE_CODING USB_line_coding;
volatile char USB_READY = 0;
volatile char USB_EP0_RECEIVED = 0;
volatile char USB_EP1_RECEIVED = 0;
volatile char USB_EP2_RECEIVED = 0;
volatile char USB_EP0_DMA_DONE = 0;
volatile char USB_EP1_DMA_DONE = 0;
volatile char USB_EP2_DMA_DONE = 0;
volatile char USB_EJECTED;
volatile char USB_POWERED;
volatile char USB_EJECT_REQUEST;

unsigned char __attribute__ ((coherent, aligned(8))) USB_data_buffer[USB_DATA_BUFFER_SIZE];
unsigned char __attribute__ ((coherent, aligned(8))) USB_EP0_buffer[USB_ENDPOINT_BUFFER_SIZE];
unsigned char __attribute__ ((coherent, aligned(8))) USB_EP1_buffer[USB_ENDPOINT_BUFFER_SIZE];
unsigned char __attribute__ ((coherent, aligned(8))) USB_EP2_buffer[USB_ENDPOINT_BUFFER_SIZE];

void USB_init_endpoints()
{
    USBE1CSR0bits.MODE = 0;     // EP1 is OUT (OUT from host = in to device = RX)
    
    // Clear all interrupt flags
	USBCSR0bits.EP0IF = 0;
    USBCSR0bits.EP1TXIF = 0;
    USBCSR1bits.EP1RXIF = 0;
    USBCSR0bits.EP2TXIF = 0;
    USBCSR1bits.EP2RXIF = 0;
	
    // Set the maximum transfer length for each endpoint
	// Configure endpoint 0 first.
    USBE0CSR0bits.TXMAXP = 64; // Set endpoint 0 buffer size to 16 bytes (can be a maximum of 64 for EP0)

    // And next my custom endpoints
    USBE1CSR0bits.TXMAXP = 512;   // Endpoint 1 - Maximum TX payload / transaction - 512
    USBE1CSR1bits.RXMAXP = 512;   // Endpoint 1 - Maximum TX payload / transaction - 512

    USBE2CSR0bits.TXMAXP = 512;   // Endpoint 2 - Maximum TX payload / transaction - 512
    USBE2CSR1bits.RXMAXP = 512;   // Endpoint 2 - Maximum TX payload / transaction - 512

	// Specify which kinds of endpoint we will be using
    USBE1CSR2bits.PROTOCOL = 2; // Endpoint 1 - Bulk mode
    USBE2CSR2bits.PROTOCOL = 2; // Endpoint 2 - Bulk mode
            
	// Enable DISNYET
    USBE1CSR1bits.PIDERR = 0; // Clear DISNYET to enable NYET replies
    USBE2CSR1bits.PIDERR = 0; // Clear DISNYET to enable NYET replies

	// Set up buffer locations for endpoint 1
    USBCSR3bits.ENDPOINT = 1;
    USBFIFOA = 0x00A000A;    
    USBIENCSR0bits.CLRDT = 1;

	// Set up buffer locations for endpoint 2
    USBCSR3bits.ENDPOINT = 2;
    USBFIFOA = 0x04A004A;    
    USBIENCSR0bits.CLRDT = 1;

    USBE1CSR3bits.TXFIFOSZ = 0x9;   // Transmit FIFO Size bits - 512 bytes
    USBE1CSR3bits.RXFIFOSZ = 0x9;   // Transmit FIFO Size bits - 512 bytes
    USBE2CSR3bits.TXFIFOSZ = 0x9;   // Transmit FIFO Size bits - 512 bytes
    USBE2CSR3bits.RXFIFOSZ = 0x9;   // Transmit FIFO Size bits - 512 bytes
	
	// Set maximum size for each packet before splitting occurs
    USBOTGbits.TXFIFOSZ = 0b0110; // 0b0110 = 512 bytes
    USBOTGbits.RXFIFOSZ = 0b0110; // 0b0110 = 512 bytes

	// Set receive endpoint 1 to High Speed
    USBE1CSR3bits.SPEED = 1;
    USBE2CSR3bits.SPEED = 1;
    
	// Disable Isochronous mode for endpoints 1 and 2
    USBE1CSR0bits.ISO = 0;      // Isochronous TX Endpoint Disable bit (Device mode).
    USBE2CSR0bits.ISO = 0;      // Isochronous TX Endpoint Disable bit (Device mode).

    // Set up endpoint interrupts
	// Initially clear all the interrupt enables (IE)
    USBCSR1 = 0;

    USBCSR1bits.EP0IE = 1;      // Endpoint 0 interrupt enable
    USBCSR1bits.EP1TXIE = 1;    // Endpoint 1 TX interrupt enable
    USBCSR2bits.EP1RXIE = 1;    // Endpoint 1 RX interrupt enable
    USBCSR1bits.EP2TXIE = 1;    // Endpoint 2 TX interrupt enable
    USBCSR2bits.EP2RXIE = 1;    // Endpoint 2 RX interrupt enable
    USBCSR2bits.RESETIE = 1;
    USBCSR2bits.RESUMEIE = 1;
    USBCSR2bits.SOFIE = 1;

	// Set current endpoint to EP2
    USBCSR3bits.ENDPOINT = 2;
}

void USB_init()
{
    USBCSR0 = 0;
    
    USB_EP0_RECEIVED = 0;
    USB_EP2_RECEIVED = 0;
    USB_init_endpoints();

    USBCSR0bits.SOFTCONN = 0;   // Initially disable it while we go into setup

    USB_address = 0;			// We haven't been given an address yet    
    USBCSR0bits.FUNC = 0;       // Initially set the USB address to 0 until later when the host assigns us an address

    USBCSR2bits.RESETIE = 1;    // Enable the reset interrupt
    IEC4bits.USBIE = 1;         // Enable the USB interrupt
    USBCRCONbits.USBIE = 1;     // Enable USB module interrupt
    IFS4bits.USBIF = 0;         // Clear the USB interrupt flag.
    IPC33bits.USBIP = 7;        // USB Interrupt Priority 7
    IPC33bits.USBIS = 1;        // USB Interrupt Sub-Priority 1
    IPC33bits.USBDMAIP = 5;
    IPC33bits.USBDMAIS = 1;
    IFS4bits.USBDMAIF = 0;
    IEC4bits.USBDMAIE = 1;

    USBCSR0bits.HSEN = 1;       // Enable High Speed (480Mbps) USB mode
    
    USB_line_coding.baud = 9600;
    USB_line_coding.data_bits = 8;
    USB_line_coding.parity_bits = 0;
    USB_line_coding.stop_bits = 0;    
}

int USB_EP0_wait_TXRDY()
{
    int timeout;
    
    timeout = 0;
    
    while (USBE0CSR0bits.TXRDY)
    {
        timeout++;
        if (timeout > USB_EP0_WAIT_TIMEOUT) return 1;
    };
    
    return 0;
}

int USB_EP1_wait_TXRDY()
{
    int timeout;
    
    timeout = 0;
    
    while (USBE1CSR0bits.TXPKTRDY)
    {
        timeout++;
        if (timeout > USB_EP0_WAIT_TIMEOUT) return 1;
    };
    
    return 0;
}

void __attribute__((vector(_USB_DMA_VECTOR), interrupt(ipl5srs), nomips16)) USB_DMA_handler()
{   
    unsigned int DMAINT;
    
    DMAINT = USBDMAINT; // This zeroes it!
    
    USBDMAINT = 0;
    
    if (DMAINT & 1 << 0) // DMA1IF
    {
        if (USBDMA1Cbits.DMADIR == 1) // RX or TX?
            USBE1CSR0bits.TXPKTRDY = 1; 
        USB_EP1_DMA_DONE = 1;
    }
    
    if (DMAINT & 1 << 1) // DMA2IF
    {
        if (USBDMA2Cbits.DMADIR == 1) // RX or TX?
            USBE2CSR0bits.TXPKTRDY = 1; 
        USB_EP2_DMA_DONE = 1;    
    }
    
    if (DMAINT & 1 << 2) // DMA3IF
    {
        if (USBDMA3Cbits.DMADIR == 1) // RX
//            USBE0CSR0bits.RXRDYC = 1;
  //      else
            USBE0CSR0bits.TXRDY = 1;        
        USB_EP0_DMA_DONE = 1;
    }
    
    IFS4bits.USBDMAIF = 0;
}

extern __inline__ unsigned int __attribute__((always_inline)) virt_to_phys(const void* p) 
{ 
 return (int)p<0?((int)p&0x1fffffffL):(unsigned int)((unsigned char*)p+0x40000000L); 
}

void USB_EP1_send_DMA(unsigned char *buffer, uint32_t dmaCount)
{
    USB_EP1_DMA_DONE = 0;

    USBE1CSR0bits.MODE = 1;     // EP1 is TX
    
    USBDMA1Abits.DMAADDR = virt_to_phys(buffer);
    USBDMA1Nbits.DMACOUNT = dmaCount;
    USBDMA1Cbits.DMABRSTM = 3;
    USBDMA1Cbits.DMAMODE = 0;
    USBDMA1Cbits.DMAEP = 1; // OUT Endpoint 1
    // DMADIR: DMA Transfer Direction bit
    // 1 = DMA Read (TX endpoint)
    // 0 = DMA Write (RX endpoint)
    USBDMA1Cbits.DMADIR = 1;
    USBDMA1Cbits.DMAIE = 1;
    // DMAEN: DMA Enable bit
    // 1 = Enable the DMA transfer and start the transfer 
    // 0 = Disable the DMA transfer
    USBDMA1Cbits.DMAEN = 1;    
}

void USB_send_EP1(unsigned char *buffer, int length)
{
    int cnt;
    unsigned char *FIFO_buffer;

    USBE1CSR0bits.MODE = 1;     // EP1 is TX
        
    FIFO_buffer = (unsigned char *)&USBFIFO1;
    
    if (USB_EP1_wait_TXRDY()) return;
    
    cnt = 0;
    
    while (cnt < length)
    {
        *FIFO_buffer = *buffer++; // Send the bytes
        cnt++;
        // Have we sent 8 bytes?
        if ((cnt > 0) && (cnt % 512 == 0))
        {
            // Set TXRDY and wait for it to be cleared before sending any more bytes
            USBE1CSR0bits.TXPKTRDY = 1;            
            if (USB_EP1_wait_TXRDY()) return;
        }
    }
    
    if ((cnt % 512) != 0) USBE1CSR0bits.TXPKTRDY = 1;            
//    USBE0CSR0bits.TXRDY = 1;            
}

void USB_EP0_send(unsigned char *buffer, uint32_t length)
{
    int cnt;
    unsigned char *FIFO_buffer;
        
    FIFO_buffer = (unsigned char *)&USBFIFO0;
    
    if (USB_EP0_wait_TXRDY()) return;
    
    cnt = 0;
    
    while (cnt < length)
    {
        *FIFO_buffer = *buffer++; // Send the bytes
        cnt++;
        // Have we sent 64 bytes?
        if ((cnt > 0) && (cnt % 64 == 0))
        {
            // Set TXRDY and wait for it to be cleared before sending any more bytes
            USBE0CSR0bits.TXRDY = 1;            
            if (USB_EP0_wait_TXRDY()) return;
        }
    }
    
    USBE0CSR0bits.TXRDY = 1;
    USB_EP0_wait_TXRDY();
    
    // An additional step is required for transfers < 4 bytes it seems
    if (length < 4)
        USBE0CSR0bits.RXRDYC = 1;
}

void USB_EP0_receive(unsigned char *buffer, uint32_t length)
{
    int cnt;
    unsigned char *FIFO_buffer;
    
    // Get 8-bit pointer to USB FIFO for endpoint 0
    FIFO_buffer = (unsigned char *)&USBFIFO0;
    
    for(cnt = 0; cnt < length; cnt++)
    {
        // Read in one byte at a time
        *buffer++ = *(FIFO_buffer + (cnt & 3));
    }
        
    USBE0CSR0bits.RXRDYC = 1;
}

void USB_EP2_receive_DMA(unsigned char *buffer, uint32_t dmaCount)
{
    USB_EP2_DMA_DONE = 0;
    USBE2CSR0bits.MODE = 0;     // EP2 is RX
    
    USBDMA2Abits.DMAADDR = virt_to_phys(buffer);
    USBDMA2Nbits.DMACOUNT = dmaCount;
    USBDMA2Cbits.DMABRSTM = 3;
    USBDMA2Cbits.DMAMODE = 0;
    USBDMA2Cbits.DMAEP = 2; // Endpoint 2
    USBDMA2Cbits.DMADIR = 0; // 0 = RX
    USBDMA2Cbits.DMAIE = 1;
    USBDMAINTbits.DMA2IF = 0;
    USBDMA2Cbits.DMAEN = 1; // Enable and start
}

int USB_receive_EP2()
{
    unsigned char *FIFO_buffer;
    int cnt;
    int bytes_received;
    
    USBE2CSR0bits.MODE = 0;     // EP2 is RX

    bytes_received = USBE2CSR2bits.RXCNT;
    
    FIFO_buffer = (unsigned char *)&USBFIFO2;
    
    for(cnt = 0; cnt < bytes_received; cnt++)
    {
        USB_EP2_buffer[cnt] = *(FIFO_buffer + (cnt & 3)); // Store the received bytes in array ep0data[].
    }
    
    USBE2CSR1bits.RXPKTRDY = 0;
    
    return bytes_received;
}

void __attribute__((vector(_USB_VECTOR), interrupt(ipl7srs), nomips16)) USB_handler()
{
    unsigned char RESETIF, EP0IF, EP1TXIF, EP1RXIF, EP2RXIF;
    unsigned int CSR0, CSR1, CSR2;
    
    CSR0 = USBCSR0;
    
    EP0IF = (CSR0 & (1<<16)) ? 1 : 0;
    EP1TXIF = (CSR0 & (1<<17)) ? 1 : 0;
    
    CSR1 = USBCSR1;
    
    EP1RXIF = (CSR1 & (1 << 1)) ? 1 : 0;
    EP2RXIF = (CSR1 & (1 << 2)) ? 1 : 0;
    
    CSR2 = USBCSR2;
    
    RESETIF = (CSR2 & (1 << 18)) ? 1 : 0;
    
    if(RESETIF)
    {
        // So when this is set, the USB bus has been reset and we actually need to set up all the USB endpoints now

        USB_init_endpoints();
        
        USBCSR2bits.RESETIF = 0;      
        
    }
    
    /* Endpoint 0 Interrupt Handler */
    if(EP0IF)
    {
        USB_EP0_RECEIVED = 1;
        USBCSR0bits.EP0IF = 0;  // Clear the USB EndPoint 0 Interrupt Flag.
    }
    
    if (EP1RXIF)
    {
        USBCSR1bits.EP1RXIF = 0;
        USB_EP1_RECEIVED = 1;
    }
    
    if(EP1TXIF)
    { 
        USBCSR0bits.EP1TXIF = 0;        // Supposedly Cleared By Hardware (Clear Just In Case).
        USBCSR3bits.ENDPOINT = 1;
        USBIENCSR0bits.MODE = 0;
    } 
        
    if (EP2RXIF)
    { 
        USBCSR1bits.EP2RXIF = 0;
        USBCSR3bits.ENDPOINT = 2;
        USBIENCSR0bits.MODE = 0;
        USB_EP2_RECEIVED = 1;
    } 

    if (USBCSR2bits.SOFIF)
    {
        USBCSR2bits.SOFIF = 0;
    }
    
    IFS4bits.USBIF = 0;
}

void USB_connect()
{
    
    USBCSR0bits.SUSPEN = 0;
    USBCSR0bits.SOFTCONN = 1;
}

void USB_disconnect()
{
    IEC4bits.USBIE = 0;
    USBCSR0bits.SUSPEN = 1;
    USBOTGbits.SESSION = 0;
    USBCSR0bits.SOFTCONN = 0;
    USB_EJECTED = 1;
}

void USB_EP0_tasks()
{
    // We're not going to bother with whether bmRequestType is IN or OUT for the most part
    switch (USB_transaction.bmRequest)
    {
        case 0xFF: // Mass storage reset
        {
            // Upon reset, we need to reset all endpoints
            USB_init_endpoints();

            break;
        }
        case 0xFE: // Get max LUN (Logical Unit)
        {
            // Send a 0
            USB_EP0_buffer[0] = 0;
            USB_EP0_send(USB_EP0_buffer, 1);
            

            break;
        }
        case 0xC: // Sync frame
        {
            USBE0CSR0bits.STALL = 1;
            break;
            
        }
        case 0x0: 
        {
            if (USB_transaction.bmRequestType == 0x80) // Get status
                USBE0CSR0bits.RXRDYC = 1;
            if (USB_transaction.bmRequestType == 0x00) // Select function
                USBE0CSR0bits.RXRDYC = 1;
            break;            
        }
        
        case 0x1: // Get report via control endpoint
        {
            // Required to support this, just send 4 0's to make it happy            
            break;
        }
        
        case 0x2: // Get idle
        {
            break;
        }
        
        case 0x3: // Get protocol
        {
            break;
        }
        
        case 0xB: // Set protocol
        {
            break;
        }
        
        case 0xA: // Set idle
        {
            USB_READY = 1;
            break;
        }
        
        case 0x5: // Set USB address
        {
            USBE0CSR0bits.RXRDYC = 1;
            USB_address = USB_EP0_buffer[2];
            USB_set_address = 1;
            USB_READY = 1;
            break;
        }

        case 0x6: // Get descriptor
        {
            switch (USB_transaction.wValue >> 8)
            {
                case 0x1: // Get device descriptor
                {
                    if (USB_transaction.wLength < USB_DEVICE_DESCRIPTION_LENGTH)
                        USB_EP0_send(USB_device_description, USB_transaction.wLength);
                    else
                        USB_EP0_send(USB_device_description, USB_DEVICE_DESCRIPTION_LENGTH);
                    break;
                }
                
                case 0x2: // Get configuration descriptor
                {
                    if (USB_transaction.wLength < USB_CONFIG_DESCRIPTOR_LENGTH)
                        USB_EP0_send(USB_config_descriptor, USB_transaction.wLength);
                    else
                        USB_EP0_send(USB_config_descriptor, USB_CONFIG_DESCRIPTOR_LENGTH);
                    break;
                }
                
                case 0x3: // Get string descriptors
                {
                    switch (USB_transaction.wValue & 0xFF)
                    {
                        case 0x0: // String 0 - Language ID
                        {
                            memcpy(USB_EP0_buffer, string0, sizeof(string0));
                            USB_EP0_send(USB_EP0_buffer, sizeof(string1));
                            break;
                        }
                        case 0x1: // String 1 - Vendor
                        {
                            memcpy(USB_EP0_buffer, string1, sizeof(string1));
                            USB_EP0_send(USB_EP0_buffer, sizeof(string1));
                            break;
                        }
                        case 0x2: // String 2 - Product name
                        {
                            memcpy(USB_EP0_buffer, string2, sizeof(string2));
                            USB_EP0_send(USB_EP0_buffer, sizeof(string2));
                            break;
                        }
                        case 0x3: // String 3 - Serial number
                        {
                            memcpy(USB_EP0_buffer, string3, sizeof(string3));
                            USB_EP0_send(USB_EP0_buffer, sizeof(string3));
                            break;
                        }
                        case 0x4: // String 4 - Configuration
                        {
                            memcpy(USB_EP0_buffer, string4, sizeof(string4));
                            USB_EP0_send(USB_EP0_buffer, sizeof(string4));
                            while (!USB_EP0_DMA_DONE);
                            break;
                        }
                        case 0x5: // String 5 - Interface
                        {
                            memcpy(USB_EP0_buffer, string5, sizeof(string5));
                            USB_EP0_send(USB_EP0_buffer, sizeof(string5));
                            break;
                        }
                        default: 
                        {
                            break;
                        }
                    }
                    
                    case 0x21: // HID HID
                    {
                        break;
                    }
                    
                    case 0x22: // HID report
                    {
                        break;                        
                    }
                }
                default: 
                {
                    break;
                }
            }
            break;
        }
        
        case 0x9: // Set report
        {
            // Zegads we have enumeration!
            USBE0CSR0bits.RXRDYC = 1; // Respond with this on EP0
            break;
        }
        
        case 0x20:
        {
            if (USB_transaction.bmRequestType == 0x21)
            {
                // Set line coding
                USB_EP0_receive(USB_EP0_buffer, USB_transaction.wLength);
                USBE0CSR0bits.RXRDYC = 1;

                
                USB_line_coding.baud = ((int)USB_EP0_buffer[3] << 24) | ((int)USB_EP0_buffer[2] << 16) | ((int)USB_EP0_buffer[1] << 8) | USB_EP0_buffer[0];
                USB_line_coding.parity_bits = USB_EP0_buffer[4];
                USB_line_coding.stop_bits = USB_EP0_buffer[5];
                USB_line_coding.data_bits = USB_EP0_buffer[6];
            }
            break;
           
        }
        
        case 0x21:
        {
            // Get line coding
            USB_EP0_buffer[0] = USB_line_coding.baud & 0xFF;
            USB_EP0_buffer[1] = (USB_line_coding.baud >> 8) & 0xFF;
            USB_EP0_buffer[2] = (USB_line_coding.baud >> 16) & 0xFF;
            USB_EP0_buffer[3] = USB_line_coding.baud >> 24;
            USB_EP0_buffer[4] = USB_line_coding.stop_bits;
            USB_EP0_buffer[5] = USB_line_coding.parity_bits;
            USB_EP0_buffer[6] = USB_line_coding.data_bits;
            USB_EP0_send(USB_EP0_buffer, 7);
            break;
        }
        
        
        default: 
        {
            USBE0CSR0bits.STALL = 1;
            break;
        }
    }    
}

void USB_tasks()
{    
    if (USB_EP0_RECEIVED)
    {
        USB_EP0_RECEIVED = 0;
        
        if (USB_EJECT_REQUEST)
        {
            USB_EJECT_REQUEST = 0;
            // DIsconnect the hardware now
            USB_EJECTED = 1;
            return;
        }
        
        if (USB_set_address) // Do we need the set the USB address?
        {
            USBCSR0bits.FUNC = USB_address & 0x7F;
            USB_set_address = 0;
        }

        USB_EP0_receive(USB_EP0_buffer, USBE0CSR2bits.RXCNT);

        USB_transaction.bmRequestType = USB_EP0_buffer[0];
        USB_transaction.bmRequest = USB_EP0_buffer[1];
        USB_transaction.signature = ((int)USB_EP0_buffer[0] << 8) | USB_EP0_buffer[1];
        USB_transaction.wValue = (int)(USB_EP0_buffer[3] << 8) | USB_EP0_buffer[2];
        USB_transaction.wIndex = (int)(USB_EP0_buffer[5] << 8) | USB_EP0_buffer[4];
        USB_transaction.wLength = (int)(USB_EP0_buffer[7] << 8) | USB_EP0_buffer[6];

        USB_EP0_tasks();

        if (USB_transaction.wLength == 0) USBE0CSR0bits.DATAEND = 1; // End of Data Control bit (Device mode) 
                
        if (USBE0CSR0bits.SETEND) 
        {
            USBE0CSR0bits.SETENDC = 1;
        }        
    }

    if (USB_EP1_RECEIVED)
    {
        USB_EP1_RECEIVED = 0;
        USB_EP1_tasks();
    }
    
   // Handle EP2
    if (USB_EP2_RECEIVED)
    {
        USB_EP2_RECEIVED = 0;
        USB_EP2_tasks();
    }
}
