/*
 * USB_MSD.c by Aidan Mocke (29/12/2020)
 * -------------------------------------
 * Works in conjunction with USB.c and mmcPIC32.c
 * Contains all the functions needed to use an attached SD card as a Mass Storage Device
 *  
 */
#include "USB.h"
#include <xc.h>
#include "ff.h"
#include "mmcpic32.h"
#include <string.h>

char USB_DEVICE_DESCRIPTION_LENGTH = 18;
char USB_CONFIG_DESCRIPTOR_LENGTH = 32;

unsigned char USB_device_description[] = {
/* Descriptor Length						*/ 0x12, /* Always 18 or 0x12 */
/* DescriptorType: DEVICE					*/ 0x01,
/* bcdUSB (ver 2.0)							*/ 0x00,0x02,
/* bDeviceClass								*/ 0x00,
/* bDeviceSubClass							*/ 0x00,
/* bDeviceProtocol							*/ 0x00,
/* bMaxPacketSize0							*/ 0x40, /* Always 64 or 0x40 for High Speed USB */
/* idVendor									*/ 0xD8,0x04, /* e.g. - 0x04D8 - Microchip VID */
/* idProduct								*/ 0x09,0x00, /* e.g. - 0x0000 */
/* bcdDevice								*/ 0x00,0x01, /* e.g. - 01.00 */
/* iManufacturer							*/ 0x01,
/* iProduct									*/ 0x02,
/* iSerialNumber							*/ 0x03,
/* bNumConfigurations						*/ 0x01
};

unsigned char USB_config_descriptor[] = {
/*********************************************************************
 Configuration Descriptors 
 *********************************************************************/
/*  bLength (Descriptor Length)             */ 9, 
/*  bDescriptorType: CONFIG					*/ 0x02,
/*  wTotalLength							*/ 0x20,0x00,
/*  bNumInterfaces							*/ 1,
/*  bConfigurationValue						*/ 1, 
/*  iConfiguration							*/ 0, 
/*  bmAttributes							*/ 0x80, /* bit 6 set = bus powered = 0x80, 0xC0 is for self powered */
/*  bMaxPower								*/ 0x32, /* Value x 2mA, set to 0 for self powered */
/*********************************************************************
  Interface 0 - HID Class
 *********************************************************************/                                               
/* bLength                                  */ 0x09,
/* bDescriptorType: INTERFACE               */ 0x04,
/* bInterfaceNumber                         */ 0x00,
/* bAlternateSetting                        */ 0x00,
/* bNumEndpoints: 2 endpoint(s)             */ 0x02,
/* bInterfaceClass: Mass Storage            */ 0x08,
/* bInterfaceSubclass:SCSI transparent      */ 0x06,
/* bInterfaceProtocol:Bulk only             */ 0x50,
/* iInterface                               */ 0x00,
/*********************************************************************
Endpoint 1 (Bulk IN)
 *********************************************************************/                                               
/* bLength                                  */ 0x07,
/* bDescriptorType: ENDPOINT                */ 0x05,
/* bEndpointAddress: IN Endpoint 1          */ 0x81, 
/* bmAttributes: BULK                       */ 0x02,
/* max packet size (LSB)                    */ 0x00,
/* max packet size (MSB)                    */ 0x02,
/* polling interval                         */ 0x00,
/*********************************************************************
 Endpoint 1 (Bulk OUT)
 *********************************************************************/
/* bLength                                  */ 0x07,
/* bDescriptorType: ENDPOINT                */ 0x05,
/* bEndpointAddress: OUT Endpoint 2         */ 0x02,
/* bmAttributes: BULK                       */ 0x02,
/* max packet size (LSB)                    */ 0x00,
/* max packet size (MSB)                    */ 0x02,
/* bInterval: None for BULK                 */ 0x00,
};

static unsigned char inquiryAnswer[36] = {
0x00, // Peripheral Device Type (PDT) - SBC Direct-access device
0x80, // Removable Medium Bit is Set
0x04, // Version
0x02, // Obsolete[7:6],NORMACA[5],HISUP[4],Response Data Format[3:0]
0x1f, // Additional Length
0x00, // SCCS[7],ACC[6],TPGS[5:4],3PC[3],Reserved[2:1],PROTECT[0]
0x00, // BQUE[7],ENCSERV[6],VS[5],MULTIP[4],MCHNGR[3],Obsolete[2:1],ADDR16[0]
0x00, // Obsolete[7:6],WBUS116[5],SYNC[4],LINKED[3],Obsolete[2],CMDQUE[1],VS[0]
'A', 'i', 'd', 'a', 'n', ' ', ' ', ' ', // Vendor Identification
'M', 'a', 's', 's', ' ', 'S', 't', 'o', 'r', 'a', 'g', 'e', ' ', ' ', ' ', ' ', // Product Identification
'0', '.', '0', '1' // Product Revision Level
};

static unsigned char modeSenseAnswer[4] = {
0x03, 0x00, 0x00, 0x00
};

//Reference: http://www.usb.org/developers/docs/devclass_docs/usb_msc_boot_1.0.pdf
static unsigned char requestSenseAnswer[18] = {
0x70, // Error Code
0x00, // Segment Number (Reserved)
0x02, // ILI, Sense Key
0x00, 0x00, 0x00, 0x00, // Information
0x0A, // Additional Sense Length (n-7), i.e. 17-7
0x00, 0x00, 0x00, 0x00, // Command Specific Information
0x3A, // Additional Sense Code
0x00, // Additional Sense Qualifier (optional)
0x00, 0x00, 0x00, 0x00 // Reserved
};

int read_capacity[8];

unsigned char MSD_ready = 1;

USB_CBW usb_cbw;
USB_CSW usb_csw;

unsigned long int SD_disk_size;

void USB_EP1_tasks()
{
    
}

void USB_EP2_tasks()
{
    int cnt;
    USB_CBW cbw;
    USB_CSW csw;
    int operation_code;
    int read_length;
    int read_address;
    DRESULT result;
    unsigned char *addr;
    unsigned int capacity_reversed;
    char toggle;
    int blocks_left;
    int blocks_to_read;
    
    if (USB_receive_EP2() != 31) return; // CBWs are always 31 bytes in length

    // Copy the CBW into a CBW struct
    memcpy(&cbw, USB_EP2_buffer, 31);

    cbw.bCBWLUN &= 0x0F; // Only the lower 4 bits matter

    csw.bCSWStatus = 0; // Good
    csw.dCSWDataResidue = 0; // No residue - what a word
    csw.dCSWSignature = 0x53425355;
    csw.dCSWTag = cbw.dCBWTag;

    if (cbw.dCBWSignature != 0x43425355) return; // Signature must always be 0x43425355 for CBWs

    operation_code = cbw.CBWCB[0]; // Actual command number

    switch (operation_code)
    {
        case 0x0: // Test unit ready
        {
            csw.bCSWStatus = (MSD_ready != 1); // If ready, return 0 for no error
            if (!MSD_ready) 
            {
                requestSenseAnswer[2] = 2;
                csw.bCSWStatus |= 1; // Error occurred
            }
            break;
        }
        case 0x3: // Request sense CMD
        {
            USB_send_EP1(requestSenseAnswer, sizeof(requestSenseAnswer));

            // Reset the sense data now
            if (MSD_ready)
                requestSenseAnswer[2] = 0;
            else
                requestSenseAnswer[2] = 2;
            break;
        }
        case 0x12: // Inquiry
        {
            // Send off inquiryAnswer
            USB_send_EP1(inquiryAnswer, sizeof(inquiryAnswer));
            break;
        }
        case 0x1A: // Mode sense (6)
        {
            // For write protect, set modeSenseAnswer[2] |= 0x80;
            USB_send_EP1(modeSenseAnswer, sizeof(modeSenseAnswer));
            break;
        }
        case 0x1B: // Start/stop unit
        {
            MSD_ready = ((cbw.CBWCB[4] >> 1) == 0);
            USB_EJECT_REQUEST = (MSD_ready == 0);
            USBE0CSR0bits.RXRDYC = 1;
            break;
        }
        case 0x1E: // Prevent allow removal
        {
            // NO, we DO NOT want this otherwise Windows enables write caching!
            csw.bCSWStatus = 1;
            requestSenseAnswer[2] = 0x5; // Illegal command
            break;
        }
        case 0x23: // Read format capacity
        {
            if (MSD_ready)
            {
                capacity_reversed = SD_disk_size >> 24; // High byte now low
                capacity_reversed |= (SD_disk_size & 0x00FF0000) >> 8; // Next byte
                capacity_reversed |= (SD_disk_size & 0x0000FF00) << 8; // Next byte
                capacity_reversed |= (SD_disk_size & 0x000000FF) << 25; // Low byte now high
                read_capacity[0] = 0x08000000; // Capacity list length - Number of descriptors = 1 x 8 = 8
                read_capacity[1] = capacity_reversed; // Number of sectors
                read_capacity[2] = 0x02020000; // Size of sector, OR 2
                memcpy(USB_EP1_buffer, read_capacity, 12);
                USB_send_EP1(USB_EP1_buffer, 12);
            }
            else
            {
                csw.bCSWStatus |= 1; // Error occurred   
                requestSenseAnswer[2] = 2;
            }

            break;
        }
        case 0x25: // Read capacity (10))
        {
            if (MSD_ready)
            {
                SD_disk_size -= 1;
                capacity_reversed = SD_disk_size >> 24; // High byte now low
                capacity_reversed |= (SD_disk_size & 0x00FF0000) >> 8; // Next byte
                capacity_reversed |= (SD_disk_size & 0x0000FF00) << 8; // Next byte
                capacity_reversed |= (SD_disk_size & 0x000000FF) << 25; // Low byte now high
                SD_disk_size += 1;
                read_capacity[0] = capacity_reversed; // Number of blocks Backwards, yay
                read_capacity[1] = 0x00020000; // Block size in bytes (512) Backwards, yay
                memcpy(USB_EP1_buffer, read_capacity, 8);
                USB_send_EP1(USB_EP1_buffer, 8);
            }
            else
            {
                csw.bCSWStatus |= 1; // Error occurred   
            }
            break;
        }
        case 0x28: // Read (10)
        {
            if (MSD_ready)
            {
                read_length = (int)(cbw.CBWCB[7] << 8) | cbw.CBWCB[8];
                read_address  = (int)(cbw.CBWCB[2] << 24) | (int)(cbw.CBWCB[3] << 16) | (cbw.CBWCB[4] << 8) | (cbw.CBWCB[5]);

                blocks_left = read_length;
                toggle = 0;

                // Start SD multiblock read
                result = SD_start_multi_block_read(read_address);
                // Read first block
                addr = USB_data_buffer;
                result |= SD_read_multi_block(addr, 512);
                blocks_left--;

                // Wait until we are ready to send on EP1
                USB_EP1_wait_TXRDY();

                while (blocks_left)
                {
                    // Send this block via DMA
                    USB_EP1_DMA_DONE = 0;
                    addr = USB_data_buffer + (toggle * 512);
                    USB_EP1_send_DMA(addr, 512);

                    // While it's sending, read the next block in
                    toggle ^= 1;
                    addr = USB_data_buffer + (toggle * 512);
                    result |= SD_read_multi_block(addr, 512);

                    // Wait until DMA transfer is done before continuing
                    while (!USB_EP1_DMA_DONE);
                    USB_EP1_wait_TXRDY();
                    blocks_left--;
                }

                // Stop the SD card multiblock read
                SD_stop_multi_block_read();

                // There will be one block that is left unsent, send it now
                USB_EP1_DMA_DONE = 0;
                USB_EP1_send_DMA(addr, 512);
                while (!USB_EP1_DMA_DONE);
                USB_EP1_wait_TXRDY();

                if (result != RES_OK) 
                {
                    requestSenseAnswer[2] = 0x5;
                    requestSenseAnswer[12] = 0x20;
                    csw.dCSWDataResidue = 252;
                    csw.bCSWStatus |= 1; // Error occurred   
                }
            }
            else
            {
                csw.bCSWStatus |= 1; // Error occurred   
                requestSenseAnswer[2] = 2;
            }

            break;
        }
        case 0x2A: // Write (10)
        {
            if (MSD_ready)
            {
                read_length = (int)(cbw.CBWCB[7] << 8) | cbw.CBWCB[8];
                read_address  = (int)(cbw.CBWCB[2] << 24) | (int)(cbw.CBWCB[3] << 16) | (cbw.CBWCB[4] << 8) | (cbw.CBWCB[5]);

                USBCSR3bits.ENDPOINT = 2;

                blocks_left = read_length;

                while (blocks_left > 0)
                {
                    if (blocks_left > USB_DATA_BUFFER_SIZE / 512)
                        blocks_to_read = USB_DATA_BUFFER_SIZE / 512;
                    else
                        blocks_to_read = blocks_left;

                    for (cnt = 0; cnt < blocks_to_read; cnt++)
                    {
                        while (!USBE2CSR1bits.RXPKTRDY)
                        addr = USB_data_buffer + (512 * cnt);
                        USB_EP2_DMA_DONE = 0;
                        USB_EP2_receive_DMA(addr, 512);
                        while (!USB_EP2_DMA_DONE);
                        USBE2CSR1bits.RXPKTRDY = 0;
                        blocks_left--;
                    }                    

                    result = disk_write(0, USB_data_buffer, read_address, blocks_to_read);
                    result = RES_OK;

                    read_address += (blocks_to_read * 512);

                    if (result != RES_OK) csw.bCSWStatus |= 1; // Error occurred   
                }
            }
            else
            {
                csw.bCSWStatus |= 1; // Error occurred   
                requestSenseAnswer[2] = 2;
            }

            break;

        }

        case 0x2F: // Verify
        {
            break;
        }
        default:
        {
            requestSenseAnswer[2] = 0x5; // Illegal request (aka not supported request)
            requestSenseAnswer[12] = 0x20;
            csw.bCSWStatus = 1;

            break;
        }
    }

    // Send the CSW
    memcpy(USB_EP1_buffer, &csw, 13);
    USB_send_EP1(USB_EP1_buffer, 13);    
}

void USB_MSD_init()
{
    disk_ioctl(0, GET_SECTOR_COUNT, &SD_disk_size);    
}