/*------------------------------------------------------------------------/
/  MMCv3/SDv1/SDv2 (in SPI mode) control module
/-------------------------------------------------------------------------/
/
/  Copyright (C) 2010, ChaN, all right reserved.
/
/  This software is a free software and there is NO WARRANTY.
/  No restriction on use. You can use, modify and redistribute it for
/  personal, non-profit or commercial products UNDER YOUR RESPONSIBILITY.
/  Redistributions of source code must retain the above copyright notice.
/
/-------------------------------------------------------------------------*/

/*
 * Edited by A. Morrison to function on PIC32.
 * Ported by Riccardo Leonardi to PIC32MX795F512L  (22/11/2011)
 * Many thanks to Aiden Morrison's good work
 * Changes: parametrization of SPI port number
 * Modified by Bryn Thomas (11/09/2016) to use Enhanced Buffer SPI mode and boost read performance with 32-bit transfers
 * Ported by Aidan Mocke (04/09/2018) to work on the PIC32MZ series
 * Modified by Aidan Mocke (14/01/2019) to use #defines to make config easier
 * Special thanks to Bryn Thomas and Ivo Colleoni for their help with the #defines
 * Configuration settings moved to mmcpic32.h, which is diskio.h renamed
 * Modified by Aidan Mocke (29/12/2020) to add much faster disk writing using Enhanced Buffer SPI mode
 * DMA code removed due to inability to get it to work reliably with USB DMA
 * 
 * PLEASE NOTE:
 * 
 * You must change CS_PORT, CS_PIN, SDO_PORT, SDO_PIN and SPI_CHANNEL in mmcpic32.h to match your configuration!
*/

#include <xc.h>
#include "mmcpic32.h"
#include "integer.h"

// Send [dat] over SPI and return the reply
static BYTE xchg_spi (BYTE dat)
{
	SPI_CHAN_FUNC(BUF) = dat;
	while (!SPI_CHAN_FUNC(STATbits).RXBUFELM);
	return (BYTE)SPI_CHAN_FUNC(BUF);
}

// Wait for card to be ready
static BYTE wait_ready (void)
{
	BYTE res;

	rcvr_spi();
	do
    {
		res = rcvr_spi();
    }
	while ((res != 0xFF));

	return res;
}

// Deselect the card and release SPI bus
static void deselect (void)
{
	CS_HIGH();
	rcvr_spi();
}


// Select the card and wait until it is ready
// Returns 1: Successful, 0: Error
static int select (void)	
{
	CS_LOW();
	if (wait_ready() != 0xFF) 
    {
		deselect();
		return 0;
	}
	return 1;
}

static BYTE send_cmd (
	BYTE cmd,		// Command byte 
	DWORD arg		// Argument 
)
{
	BYTE n, res;


	if (cmd & 0x80) {	// ACMD<n> is the command sequence of CMD55-CMD<n> 
		cmd &= 0x7F;
		res = send_cmd(CMD55, 0);
		if (res > 1) return res;
	}

	// Select the card and wait for ready 
	deselect();
	if (!select()) return 0xFF;

	// Send command packet 
	xmit_spi(0x40 | cmd);			// Start + Command index 
	xmit_spi((BYTE)(arg >> 24));	// Argument[31..24] 
	xmit_spi((BYTE)(arg >> 16));	// Argument[23..16] 
	xmit_spi((BYTE)(arg >> 8));		// Argument[15..8] 
	xmit_spi((BYTE)arg);			// Argument[7..0] 
	n = 0x01;						// Dummy CRC + Stop 
	if (cmd == CMD0) n = 0x95;		// Valid CRC for CMD0(0) 
	if (cmd == CMD8) n = 0x87;		// Valid CRC for CMD8(0x1AA) 
	xmit_spi(n);

	// Receive command response 
	if (cmd == CMD12) rcvr_spi();	// Skip a stuff byte when stop reading 
	n = 10;							// Wait for a valid response in timeout of 10 attempts 
	do
		res = rcvr_spi();
	while ((res & 0x80) && --n);

	return res;			// Return with the response value 
}

// When the target system does not support socket power control, there is nothing to do in these functions and chk_power always returns 1.
// I've used it to set the CS pin to 0 (selected), set up the SPI peripheral and initialise some DMA-related stuff
static void power_on (void)
{
	CS_SETOUT();
    SPI_CHAN_FUNC(CONbits).ON = 0;          // Turn off SPI channel
    SPI_CHAN_FUNC(CONbits).MSTEN = 1;       // Enable Master Mode
    SPI_CHAN_FUNC(CONbits).CKP = 1;         // Clock Polarity Select is high (Idle = high, active = low)
    SPI_CHAN_FUNC(CONbits).CKE = 0;         // Clock Edge Select is low, transfer out SDO on idle->active edge
    SPI_CHAN_FUNC(CONbits).SMP = 1;         // Data is sampled at the end of the data output
    SPI_CHAN_FUNC(CONbits).MODE16 = 0;      // Not 16-bit mode
    SPI_CHAN_FUNC(CONbits).MODE32 = 0;      // Not 32-bit mode
    SPI_CHAN_FUNC(CONbits).SRXISEL = 0b01;  // Set SRXISEL to generate an interrupt when the receive buffer is not empty    
    SPI_CHAN_FUNC(CONbits).STXISEL = 0b01;  // Set STXISEL to generate an interrupt when the buffer is empty
    SPI_CHAN_FUNC(CONbits).ENHBUF = 1;      // Enable Enhanced Buffers
	SPI_CHAN_FUNC(CONbits).ON = 1;          // Turn on this SPI channel
}

// Power off the SD card
static void power_off (void)
{
	select();			// Wait for card to be ready 
	deselect();         // De-select the card
	SPI_CHAN_FUNC(CONbits).ON = 0;  // Disable SPI 
	Stat |= STA_NOINIT;	// Set STA_NOINIT 
}


#if _READONLY == 0
static
int xmit_datablock (	/* 1:OK, 0:Failed */
	const BYTE *buff,	/* 512 byte data block to be transmitted */
	BYTE token,			/* Data token */
    UINT count
)
{
	static BYTE resp;
	UINT bc = 512;


	while (SPI_CHAN_FUNC(STATbits).RXBUFELM > 0)
    {
        resp = SPI_CHAN_FUNC(BUF);
        bc = resp;
    }
    
    if (wait_ready() != 0xFF) return 0;
	
    xmit_spi(token);		/* Xmit a token */
	if (token != 0xFD) 
    {	/* Not StopTran token */
		
        // Send 16 bytes first
        for (bc = 0; bc < 16; bc++)
        {
            SPI_CHAN_FUNC(BUF) = *buff++;
        }
        
        bc = 512 - 16;
        
        while (bc > 0)
        {
            if (SPI_CHAN_FUNC(STATbits).RXBUFELM > 0)
            {
                // Read an 0xFF
                resp = SPI_CHAN_FUNC(BUF);
                // Send next byte
                SPI_CHAN_FUNC(BUF) = *buff++;
                bc--;
            }
        }
        
        
//        while ((SPI_CHAN_FUNC(STATbits).TXBUFELM > 0));
        
        while (SPI_CHAN_FUNC(STATbits).SPIBUSY);

        
        // Clear out the FIFO
        
        while (SPI_CHAN_FUNC(STATbits).RXBUFELM > 0)
        {
            resp = SPI_CHAN_FUNC(BUF);
        }
        
        // Get 2x CRC and 1x response
        SPI_CHAN_FUNC(BUF) = 0xFF;
        SPI_CHAN_FUNC(BUF) = 0xFF;
        SPI_CHAN_FUNC(BUF) = 0xFF;
        while (SPI_CHAN_FUNC(STATbits).SPIBUSY);
        while ((SPI_CHAN_FUNC(STATbits).RXBUFELM > 0))
            resp = SPI_CHAN_FUNC(BUF);

//		resp = rcvr_spi();			/* Receive a data response */
		if ((resp & 0x1F) != 0x05)	/* If not accepted, return with error */
			return 0;
	}


	return 1;
}
#endif	/* _READONLY */


#if _READONLY == 0
DRESULT disk_write (BYTE pdrv, const BYTE* buff, DWORD sector, UINT count)
{
	if (pdrv || !count) return RES_PARERR;
	if (Stat & STA_NOINIT) return RES_NOTRDY;
	if (Stat & STA_PROTECT) return RES_WRPRT;

	if (!(CardType & CT_BLOCK)) sector *= 512;	// Convert to byte address if needed 

	if (count == 1) {		// Single block write 
		if ((send_cmd(CMD24, sector) == 0)	// WRITE_BLOCK 
			&& xmit_datablock(buff, 0xFE, 1))
			count = 0;
	}
	else {				// Multiple block write 
		if (CardType & CT_SDC) send_cmd(ACMD23, count);
		if (send_cmd(CMD25, sector) == 0) 
        {	// WRITE_MULTIPLE_BLOCK 
            // Write all blocks with token 0xFC
            // Write call xmit_datablock again but with token 0xFD
//            if (!xmit_datablock(buff, 0xFC, count)) count = 1;
  //          if (!xmit_datablock(buff, 0xFD, 1)) count = 1;
			do {
				if (!xmit_datablock(buff, 0xFC, 1)) break;
				buff += 512;
			} while (--count);
			if (!xmit_datablock(0, 0xFD, 1))	// STOP_TRAN token 
				count = 1;
		}
	}
	deselect();

	return count ? RES_ERROR : RES_OK;
}
#endif // _READONLY 

// Receive a single 512-byte data block from the SD card
static int rcvr_datablock (	/* 1:OK, 0:Failed */
	BYTE *buff,			/* Data buffer to store received data */
	UINT btr			/* Byte count (must be multiple of 4) */
)
{
	BYTE token;


	do {							/* Wait for data packet in timeout of 100ms */
		token = rcvr_spi();
	} while (token == 0xFF);

	if(token != 0xFE) return 0;		/* If not valid data token, retutn with error */

	do {							/* Receive the data block into buffer */
		rcvr_spi_m(buff++);
		rcvr_spi_m(buff++);
		rcvr_spi_m(buff++);
		rcvr_spi_m(buff++);
	} while (btr -= 4);
	rcvr_spi();						/* Discard CRC */
	rcvr_spi();

	return 1;						/* Return with success */
}

static int rcvr_datablock_multiple(BYTE *buff, INT btr)
{
    unsigned char READ_DONE = 0;
    int sectors_left;
    int cnt;
    unsigned char crc;

    //asm("di");
    sectors_left = btr >> 9;
        
    READ_DONE = 0;

    while (!READ_DONE)
    {
        // Wait until the 0xFE token is received
        SPI_CHAN_FUNC(BUF) = 0xFF;
        SPI_CHAN_FUNC(BUF) = 0xFF;
        while (!SPI_CHAN_FUNC(STATbits).RXBUFELM);
        crc = SPI_CHAN_FUNC(BUF);        
        
        while (crc != 0xFE)
        {
            SPI_CHAN_FUNC(BUF) = 0xFF;
            while (!SPI_CHAN_FUNC(STATbits).RXBUFELM);
            crc = SPI_CHAN_FUNC(BUF);
        }
        
        while (SPI_CHAN_FUNC(STATbits).SPIBUSY);
        
        crc = SPI_CHAN_FUNC(STATbits).RXBUFELM;
        
        // Stuff buffer with 8 x  0xFFs first
        for (cnt = 0; cnt < 8; cnt++)
            SPI_CHAN_FUNC(BUF) = 0xFF;

        cnt = 512 - 8 - crc + 2; // 2 bytes of CRC
        
        while (cnt > 0)
        {
            if (SPI_CHAN_FUNC(STATbits).RXBUFELM > 0)
            {
                // Retrieve the byte
                *buff++ = SPI_CHAN_FUNC(BUF);
                cnt--;
                // Stick in another 0xFF
                SPI_CHAN_FUNC(BUF) = 0xFF;
            }
        }
               
        while (SPI_CHAN_FUNC(STATbits).SPIBUSY);
        
        while ((SPI_CHAN_FUNC(STATbits).RXBUFELM > 2))
        {
            *buff++ = SPI_CHAN_FUNC(BUF);
        }
        
        // Read the CRC
        crc = SPI_CHAN_FUNC(BUF);
        crc = SPI_CHAN_FUNC(BUF);
            
        // Sector read complete
        sectors_left--;

        // Done reading all sectors?
        if (sectors_left == 0)
            READ_DONE = 1;
        
    }
    
    //asm("ei");
    return 1;						
}

/*-----------------------------------------------------------------------*/
/* Send a data packet to MMC                                             */
/*-----------------------------------------------------------------------*/



/*-----------------------------------------------------------------------*/
/* Send a command packet to MMC                                          */
/*-----------------------------------------------------------------------*/

/*--------------------------------------------------------------------------

   Public Functions

---------------------------------------------------------------------------*/


/*-----------------------------------------------------------------------*/
/* Initialize Disk Drive                                                 */
/*-----------------------------------------------------------------------*/

DSTATUS disk_initialize (
	BYTE drv		/* Physical drive nmuber (0) */
)
{
	BYTE n, cmd, ty, ocr[4];


	power_on();							/* Force socket power on */
    FCLK_SLOW();
	for (n = 80; n; n--) rcvr_spi();	/* 80 dummy clocks */

	ty = 0;
	if (send_cmd(CMD0, 0) == 1) {			/* Enter Idle state */
		if (send_cmd(CMD8, 0x1AA) == 1) {	/* SDv2? */
			for (n = 0; n < 4; n++) ocr[n] = rcvr_spi();			/* Get trailing return value of R7 resp */
			if (ocr[2] == 0x01 && ocr[3] == 0xAA) {				/* The card can work at vdd range of 2.7-3.6V */
				while (send_cmd(ACMD41, 0x40000000));	/* Wait for leaving idle state (ACMD41 with HCS bit) */
				if (send_cmd(CMD58, 0) == 0) {			/* Check CCS bit in the OCR */
					for (n = 0; n < 4; n++) ocr[n] = rcvr_spi();
					ty = (ocr[0] & 0x40) ? CT_SD2|CT_BLOCK : CT_SD2;	/* SDv2 */
				}
			}
		} else {							/* SDv1 or MMCv3 */
			if (send_cmd(ACMD41, 0) <= 1) 	{
				ty = CT_SD1; cmd = ACMD41;	/* SDv1 */
			} else {
				ty = CT_MMC; cmd = CMD1;	/* MMCv3 */
			}
			while (send_cmd(cmd, 0));		/* Wait for leaving idle state */
			if (send_cmd(CMD16, 512) != 0)	/* Set read/write block length to 512 */
				ty = 0;
		}
	}
    
	CardType = ty;
	deselect();

	if (ty) {			/* Initialization succeded */
		Stat &= ~STA_NOINIT;	/* Clear STA_NOINIT */
		FCLK_FAST();
	} else {			/* Initialization failed */
		power_off();
	}

	return Stat;
}



/*-----------------------------------------------------------------------*/
/* Get Disk Status                                                       */
/*-----------------------------------------------------------------------*/

DSTATUS disk_status (
	BYTE drv		/* Physical drive nmuber (0) */
)
{
	if (drv) return STA_NOINIT;		/* Supports only single drive */
	return Stat;
}



/*-----------------------------------------------------------------------*/
/* Read Sector(s)                                                        */
/*-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*/
/* Read Sector(s)                                                        */
/*-----------------------------------------------------------------------*/


DRESULT disk_read (BYTE pdrv, BYTE* buff, DWORD sector, UINT count)
{
	if (pdrv || !count) return RES_PARERR;
	if (Stat & STA_NOINIT) return RES_NOTRDY;
       
	if (!(CardType & CT_BLOCK)) sector *= 512;	/* Convert to byte address if needed */

	if (count == 1) {		/* Single block read */
		if ((send_cmd(CMD17, sector) == 0)	/* READ_SINGLE_BLOCK */
			&& rcvr_datablock_multiple(buff, 512))
			count = 0;
	}
	else {				/* Multiple block read */
		if (send_cmd(CMD18, sector) == 0) {	/* READ_MULTIPLE_BLOCK */
			if (rcvr_datablock_multiple(buff, count*512)) count = 0;
			send_cmd(CMD12, 0);				/* STOP_TRANSMISSION */
		}
	}

	return count ? RES_ERROR : RES_OK;
}

DRESULT SD_start_multi_block_read(DWORD sector)
{
	if (!(CardType & CT_BLOCK)) sector *= 512;	/* Convert to byte address if needed */    
    
    if (send_cmd(CMD18, sector) == 0)
        return RES_OK;
    else
        return RES_ERROR;        
}

DRESULT SD_read_multi_block(BYTE* buff, UINT cnt)
{
    if (rcvr_datablock_multiple(buff, cnt)) 
        return RES_OK;
    else
        return RES_ERROR;
}

DRESULT SD_stop_multi_block_read()
{
    if (send_cmd(CMD12, 0) == 0)				/* STOP_TRANSMISSION */    
        return RES_OK;
    else
        return RES_ERROR;        
}


/*-----------------------------------------------------------------------*/
/* Miscellaneous Functions                                               */
/*-----------------------------------------------------------------------*/

DRESULT disk_ioctl (
	BYTE drv,		/* Physical drive nmuber (0) */
	BYTE ctrl,		/* Control code */
	void *buff		/* Buffer to send/receive data block */
)
{
	DRESULT res;
	BYTE n, csd[16], *ptr = buff;
	DWORD csize;


	if (drv) return RES_PARERR;
	if (Stat & STA_NOINIT) return RES_NOTRDY;

	res = RES_ERROR;
	switch (ctrl) {
		case CTRL_SYNC :	/* Flush dirty buffer if present */
			if (select()) {
				deselect();
				res = RES_OK;
			}
			break;

		case GET_SECTOR_COUNT :	/* Get number of sectors on the disk (WORD) */
			if ((send_cmd(CMD9, 0) == 0) && rcvr_datablock(csd, 16)) {
				if ((csd[0] >> 6) == 1) {	/* SDv2? */
					csize = csd[9] + ((WORD)csd[8] << 8) + 1;
					*(DWORD*)buff = (DWORD)csize << 10;
				} else {					/* SDv1 or MMCv2 */
					n = (csd[5] & 15) + ((csd[10] & 128) >> 7) + ((csd[9] & 3) << 1) + 2;
					csize = (csd[8] >> 6) + ((WORD)csd[7] << 2) + ((WORD)(csd[6] & 3) << 10) + 1;
					*(DWORD*)buff = (DWORD)csize << (n - 9);
				}
				res = RES_OK;
			}
			break;

		case GET_SECTOR_SIZE :	/* Get sectors on the disk (WORD) */
			*(WORD*)buff = 512;
			res = RES_OK;
			break;

		case GET_BLOCK_SIZE :	/* Get erase block size in unit of sectors (DWORD) */
			if (CardType & CT_SD2) {	/* SDv2? */
				if (send_cmd(ACMD13, 0) == 0) {		/* Read SD status */
					rcvr_spi();
					if (rcvr_datablock(csd, 16)) {				/* Read partial block */
						for (n = 64 - 16; n; n--) rcvr_spi();	/* Purge trailing data */
						*(DWORD*)buff = 16UL << (csd[10] >> 4);
						res = RES_OK;
					}
				}
			} else {					/* SDv1 or MMCv3 */
				if ((send_cmd(CMD9, 0) == 0) && rcvr_datablock(csd, 16)) {	/* Read CSD */
					if (CardType & CT_SD1) {	/* SDv1 */
						*(DWORD*)buff = (((csd[10] & 63) << 1) + ((WORD)(csd[11] & 128) >> 7) + 1) << ((csd[13] >> 6) - 1);
					} else {					/* MMCv3 */
						*(DWORD*)buff = ((WORD)((csd[10] & 124) >> 2) + 1) * (((csd[11] & 3) << 3) + ((csd[11] & 224) >> 5) + 1);
					}
					res = RES_OK;
				}
			}
			break;

		case MMC_GET_TYPE :		/* Get card type flags (1 byte) */
			*ptr = CardType;
			res = RES_OK;
			break;

		case MMC_GET_CSD :	/* Receive CSD as a data block (16 bytes) */
			if ((send_cmd(CMD9, 0) == 0)	/* READ_CSD */
				&& rcvr_datablock(buff, 16))
				res = RES_OK;
			break;

		case MMC_GET_CID :	/* Receive CID as a data block (16 bytes) */
			if ((send_cmd(CMD10, 0) == 0)	/* READ_CID */
				&& rcvr_datablock(buff, 16))
				res = RES_OK;
			break;

		case MMC_GET_OCR :	/* Receive OCR as an R3 resp (4 bytes) */
			if (send_cmd(CMD58, 0) == 0) {	/* READ_OCR */
				for (n = 0; n < 4; n++)
					*((BYTE*)buff+n) = rcvr_spi();
				res = RES_OK;
			}
			break;

		case MMC_GET_SDSTAT :	/* Receive SD statsu as a data block (64 bytes) */
			if (send_cmd(ACMD13, 0) == 0) {	/* SD_STATUS */
				rcvr_spi();
				if (rcvr_datablock(buff, 64))
					res = RES_OK;
			}
			break;

		default:
			res = RES_PARERR;
	}

	deselect();

	return res;
}



// disk_timerproc() is supposed to be a 1ms timer, I'm just keeping it in for compatibility
void disk_timerproc (void)
{
}