/* PIC32MZ_USB_SD_test - main.c - A USB MSD SD card implementation for the PIC32MZ
   by Aidan Mocke 2020-12-29

 Sources used to make this code:
 -------------------------------
 XMOS AN00125 USB Mass Storage class explanation and source - https://www.xmos.ai/download/AN00125:-USB-Mass-Storage-Device-Class(2.0.2rc1).pdf?sfw=pass1608866624
 STM32 USB source - http://en.pudn.com/Download/item/id/724481.html
 Microchip AN2554 - http://ww1.microchip.com/downloads/en/AppNotes/00002554A.pdf
 Read Format Capacities (0x23) https://www.mikrocontroller.net/attachment/41812/0x23_READFMT_070123.pdf
 AT91 ARM microcontroller USB MSD implementaion - good info on structures - http://ww1.microchip.com/downloads/en/Appnotes/doc6283.pdf
 More info on structures - https://www.renesas.com/kr/en/document/apn/sh7286-group-usb-function-module-usb-mass-storage-class?language=en
 Microchip AN1169 - http://ww1.microchip.com/downloads/en/appnotes/01169a.pdf
 MMC opcodes - http://www.procyonengineering.com/embedded/avr/avrlib/docs/html/group__mmc.html
 Silicon Labs Mass Storage Reference Design Guide - https://www.silabs.com/documents/public/application-notes/AN282.pdf
 Jan Axelson - USB mass storage (book)
 SCSI Primary Commands 3 (SPC-3) - http://www.13thmonkey.org/documentation/SCSI/spc3r23.pdf
*/

// Setup for PIC32MZ EF device for 200 MHz operation - Device Configuration Bits.
/*** DEVCFG0 ***/
#pragma config DEBUG =      OFF
#pragma config JTAGEN =     OFF
#pragma config ICESEL =     ICS_PGx1
#pragma config TRCEN =      OFF
#pragma config BOOTISA =    MIPS32
#pragma config FECCCON =    OFF_UNLOCKED
#pragma config FSLEEP =     OFF
#pragma config DBGPER =     PG_ALL
#pragma config SMCLR =      MCLR_NORM
#pragma config SOSCGAIN =   GAIN_2X
#pragma config SOSCBOOST =  ON
#pragma config POSCGAIN =   GAIN_2X
#pragma config POSCBOOST =  ON
#pragma config EJTAGBEN =   NORMAL
#pragma config CP =         OFF

/*** DEVCFG1 ***/
#pragma config FNOSC =      SPLL
#pragma config DMTINTV =    WIN_127_128
#pragma config FSOSCEN =    OFF
#pragma config IESO =       OFF
#pragma config POSCMOD =    EC
#pragma config OSCIOFNC =   OFF
#pragma config FCKSM =      CSECME
#pragma config WDTPS =      PS1048576
#pragma config WDTSPGM =    STOP
#pragma config FWDTEN =     OFF
#pragma config WINDIS =     NORMAL
#pragma config FWDTWINSZ =  WINSZ_25
#pragma config DMTCNT =     DMT31
#pragma config FDMTEN =     OFF

/*** DEVCFG2 ***/
#pragma config FPLLIDIV =   DIV_3
#pragma config FPLLRNG =    RANGE_5_10_MHZ
#pragma config FPLLICLK =   PLL_POSC
#pragma config FPLLMULT =   MUL_50
#pragma config FPLLODIV =   DIV_2
#pragma config UPLLFSEL =   FREQ_24MHZ

/*** DEVCFG3 ***/
#pragma config USERID =     0xffff
#pragma config FMIIEN =     ON
#pragma config FETHIO =     ON
#pragma config PGL1WAY =    OFF
#pragma config PMDL1WAY =   OFF
#pragma config IOL1WAY =    OFF
#pragma config FUSBIDIO =   ON

#define SYS_FREQ (200000000L)
#define VBUS_VALID 0x3

#include <xc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ff.h"
#include "mmcpic32.h"
#include "USB.h"

FIL file;                                               // file is the handle of the MP3 file we are streaming from SD card
DIR dir;                                                // Stores information about the current directory
FATFS fso;                                              // Stores information about the file system
FILINFO fileInfo;                                       // Stores information about the current file

extern void USB_MSD_init();

void set_performance_mode()
{   
	unsigned int cp0;
	
    // Unlock Sequence
    asm volatile("di"); // Disable all interrupts
    SYSKEY = 0xAA996655;
    SYSKEY = 0x556699AA;  

    // PB1DIV
    // Peripheral Bus 1 cannot be turned off, so there's no need to turn it on
    PB1DIVbits.PBDIV = 1; // Peripheral Bus 1 Clock Divisor Control (PBCLK1 is SYSCLK divided by 2)

    // PB2DIV
    PB2DIVbits.ON = 1; // Peripheral Bus 2 Output Clock Enable (Output clock is enabled)
    PB2DIVbits.PBDIV = 1; // Peripheral Bus 2 Clock Divisor Control (PBCLK2 is SYSCLK divided by 2)

    // PB3DIV
    PB3DIVbits.ON = 1; // Peripheral Bus 2 Output Clock Enable (Output clock is enabled)
    PB3DIVbits.PBDIV = 1; // Peripheral Bus 3 Clock Divisor Control (PBCLK3 is SYSCLK divided by 2)

    // PB4DIV
    PB4DIVbits.ON = 1; // Peripheral Bus 4 Output Clock Enable (Output clock is enabled)
    while (!PB4DIVbits.PBDIVRDY); // Wait until it is ready to write to
    PB4DIVbits.PBDIV = 0; // Peripheral Bus 4 Clock Divisor Control (PBCLK4 is SYSCLK divided by 1)

    // PB5DIV
    PB5DIVbits.ON = 1; // Peripheral Bus 5 Output Clock Enable (Output clock is enabled)
    PB5DIVbits.PBDIV = 1; // Peripheral Bus 5 Clock Divisor Control (PBCLK5 is SYSCLK divided by 2)

    // PB7DIV
    PB7DIVbits.ON = 1; // Peripheral Bus 7 Output Clock Enable (Output clock is enabled)
    PB7DIVbits.PBDIV = 0; // Peripheral Bus 7 Clock Divisor Control (PBCLK7 is SYSCLK divided by 1)

    // PB8DIV
    PB8DIVbits.ON = 1; // Peripheral Bus 8 Output Clock Enable (Output clock is enabled)
    PB8DIVbits.PBDIV = 1; // Peripheral Bus 8 Clock Divisor Control (PBCLK8 is SYSCLK divided by 2)

    // PRECON - Set up prefetch
    PRECONbits.PFMSECEN = 0; // Flash SEC Interrupt Enable (Do not generate an interrupt when the PFMSEC bit is set)
    PRECONbits.PREFEN = 0b11; // Predictive Prefetch Enable (Enable predictive prefetch for any address)
    PRECONbits.PFMWS = 0b010; // PFM Access Time Defined in Terms of SYSCLK Wait States (Two wait states)
    
    CFGCONbits.USBSSEN = 1;

    // Set up caching
    cp0 = _mfc0(16, 0);
    cp0 &= ~0x07;
    cp0 |= 0b011; // K0 = Cacheable, non-coherent, write-back, write allocate
    _mtc0(16, 0, cp0);  

   // Configure SRS Priority Selection
    PRISSbits.SS0 = 0;
    PRISSbits.PRI1SS = 1;
    PRISSbits.PRI2SS = 2;
    PRISSbits.PRI3SS = 3;
    PRISSbits.PRI4SS = 4;
    PRISSbits.PRI5SS = 5;
    PRISSbits.PRI6SS = 6;
    PRISSbits.PRI7SS = 7;
    
    INTCONbits.MVEC = 1; // Global interrupt enable

    // Lock Sequence
    SYSKEY = 0x33333333;
    
    asm volatile("ei"); // Enable all interrupts
}

void setup_ports()
{
    // Turn off all analog inputs, make everything digital
    ANSELA = 0;
    ANSELB = 0;
    ANSELE = 0;
    ANSELG = 0;
    
    // Make all ports outputs initially
    TRISB = 0;
    TRISC = 0;
    TRISD = 0;
    TRISE = 0;
    TRISF = 0;
    TRISG = 0;
    
    // Set all port values low initially
    LATA = 0;
    LATB = 0;
    LATC = 0;
    LATD = 0;
    LATE = 0;
    LATF = 0;
    LATG = 0;
    
    LATFbits.LATF3 = 1; // USB device mode
    CNPUFbits.CNPUF3 = 1; // Pull up to ensure device mode
    CNPUBbits.CNPUB3 = 1; // Pull up for SDI2
    CNPUBbits.CNPUB5 = 1; // Pull up for SDO2
    
    // PPS for SD card
    RPB5R = 0b0110; 		// Set SDO2 = RB5
    TRISBbits.TRISB3 = 1; 	// Set SDI2 pin to input
    SDI2R = 0b1000; 		// SDI2 = RB3
}

void delay_us(unsigned int us)
{
    
    // Convert microseconds us into how many clock ticks it will take
    us *= SYS_FREQ / 1000000 / 2; // Core Timer updates every 2 ticks

    _CP0_SET_COUNT(0); // Set Core Timer count to 0

    while (us > _CP0_GET_COUNT()); // Wait until Core Timer count reaches the number we calculated earlier
}

void delay_ms(unsigned int ms)
{
    delay_us(ms * 1000);
}


void disk_init()
{
    while(disk_initialize(0));
    f_mount(&fso, "/", 0);
    f_chdir("/");
    f_opendir(&dir, ".");    
}


int  main()
{
    set_performance_mode();
    setup_ports();
    
    TRISBSET = 1 << 4; // A button on my dev board

//    delay_ms(3000); // EC startup delay (required for using USB with EC series)
    
     // Initialise SD card
    disk_init();
    
    // Reset SPI2BRG to give us maximum speed on SPI2 (50MHz) - Works fine with all my SD cards but YMMV. Make it bigger if it doesn't work.
    SPI2BRG = 0;
    
    USB_EJECTED = 0;
    USB_POWERED = 0;
    USB_EJECT_REQUEST = 0;
    
    while (1)
    {
        while (!USB_EJECTED)
        {        
            if (USB_POWERED) USB_tasks();

            // Check if USB is connected by monitoring the VBUS voltage level
            if(USBOTGbits.VBUS == VBUS_VALID && !USB_POWERED && !USB_EJECTED)
            {
                USB_init();
                CNPUFbits.CNPUF3 = 1; // Pull up to ensure device mode
                LATFbits.LATF3 = 1; // USB device mode
                USB_connect();
                USB_POWERED = 1;
            } 
            
            // Did we lose the connection to the host?
            if(USBOTGbits.VBUS != VBUS_VALID && USB_POWERED)
            {
                USB_disconnect();
                USB_POWERED = 0;
            }
            
            // If the card has been ejected, disconnect now
            if (USB_EJECTED) 
            {
                USB_disconnect();
                USB_POWERED = 0;
            }
        }
        
        // If we get here, it means USB is not connected
        // If a button is pressed, attempt to connect it via software        
        if (PORTBbits.RB4 == 0) // Again, RB4 is a button on my devboard
        {
            USB_EJECTED = 0; // Un-eject it :)
            delay_ms(500);
            while (PORTBbits.RB4 == 0);
            MSD_ready = 1;
        }
    }
    
    return 0;
}